#include <assert.h>
#include <errno.h>
#include <netinet/in.h>
#include <poll.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <unistd.h>

#include "acceptor.h"
#include "serverpool.h"
#include "state.pb.h"

using namespace std;

extern ServerPool *servers;
extern volatile bool hupped;

int create_server_socket(int port, SocketType socket_type)
{
	int server_sock;
	if (socket_type == TCP_SOCKET) {
		server_sock = socket(PF_INET6, SOCK_STREAM, IPPROTO_TCP);
	} else {
		assert(socket_type == UDP_SOCKET);
		server_sock = socket(PF_INET6, SOCK_DGRAM, IPPROTO_UDP);
	}
	if (server_sock == -1) {
		perror("socket");
		exit(1);
	}

	int one = 1;
	if (setsockopt(server_sock, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one)) == -1) {
		perror("setsockopt(SO_REUSEADDR)");
		exit(1);
	}

	// We want dual-stack sockets. (Sorry, OpenBSD and Windows XP...)
	int zero = 0;
	if (setsockopt(server_sock, IPPROTO_IPV6, IPV6_V6ONLY, &zero, sizeof(zero)) == -1) {
		perror("setsockopt(IPV6_V6ONLY)");
		exit(1);
	}

	// Set as non-blocking, so the acceptor thread can notice that we want to shut it down.
	if (ioctl(server_sock, FIONBIO, &one) == -1) {
		perror("ioctl(FIONBIO)");
		exit(1);
	}

	sockaddr_in6 addr;
	memset(&addr, 0, sizeof(addr));
	addr.sin6_family = AF_INET6;
	addr.sin6_port = htons(port);

	if (bind(server_sock, reinterpret_cast<sockaddr *>(&addr), sizeof(addr)) == -1) {
		perror("bind");
		exit(1);
	}

	if (socket_type == TCP_SOCKET) {
		if (listen(server_sock, 128) == -1) {
			perror("listen");
			exit(1);
		}
	}

	return server_sock;
}
	
Acceptor::Acceptor(int server_sock, int port)
	: server_sock(server_sock),
	  port(port)
{
}

Acceptor::Acceptor(const AcceptorProto &serialized)
	: server_sock(serialized.server_sock()),
	  port(serialized.port())
{
}

AcceptorProto Acceptor::serialize() const
{
	AcceptorProto serialized;
	serialized.set_server_sock(server_sock);
	serialized.set_port(port);
	return serialized;
}

void Acceptor::close_socket()
{
	int ret;
	do {
		ret = close(server_sock);
	} while (ret == -1 && errno == EINTR);

	if (ret == -1) {
		perror("close");
	}
}

void Acceptor::do_work()
{
	while (!hupped) {
		// Since we are non-blocking, we need to wait for the right state first.
		// Wait up to 50 ms, then check hupped.
		pollfd pfd;
		pfd.fd = server_sock;
		pfd.events = POLLIN;

		int nfds = poll(&pfd, 1, 50);
		if (nfds == 0 || (nfds == -1 && errno == EINTR)) {
			continue;
		}
		if (nfds == -1) {
			perror("poll");
			usleep(100000);
			continue;
		}

		sockaddr_in6 addr;
		socklen_t addrlen = sizeof(addr);

		// Get a new socket.
		int sock = accept(server_sock, reinterpret_cast<sockaddr *>(&addr), &addrlen);
		if (sock == -1 && errno == EINTR) {
			continue;
		}
		if (sock == -1) {
			perror("accept");
			usleep(100000);
			continue;
		}

		// Set the socket as nonblocking.
		int one = 1;
		if (ioctl(sock, FIONBIO, &one) == -1) {
			perror("FIONBIO");
			exit(1);
		}

		// Pick a server, round-robin, and hand over the socket to it.
		servers->add_client(sock);
	}
}
